/*
 * File:	receive.c
 * 
 * Author:	Ulli Horlacher (framstag@rus.uni-stuttgart.de)
 * 
 * Contrib.:	Christian Recktenwald (chris@yeti.faveve.uni-stuttgart.de)
 * 		Rainer Bawidamann (widi@sol.wohnheim.uni-ulm.de)
 * 
 * History:	11 Aug 95   Framstag	initial version
 *              14 Aug 95   Framstag	corrected bug when receiving archives
 *              10 Sep 95   Framstag	extracted functions to spool.c
 *              31 Oct 95   Framstag	fixed security problems
 *               5 Nov 95   Framstag    added NeXT support
 *               7 Nov 95   Framstag    corrected bug when receiving to
 *                                      a non-writable directory
 * 		 2 Dec 95   Framstag	added rename option
 * 		15 Jan 96   Framstag	better error handling
 *               6 Feb 96   Framstag    added ATTRIBUTE=EXE
 *               7 Feb 96   Chris       better Convex-OS support
 *              21 Feb 96   widi        better Solaris-2 support
 *              22 Feb 96   Framstag    added bounce (forward) option
 *               6 Mar 96   Framstag    moved time stamp setting to get_date
 *              17 Mar 96   Framstag    some bug fixes
 *              27 Mar 96   Framstag    quiet mode: no questions asked
 *              28 Mar 96   Framstag    extended search for support programs
 *               1 Apr 96   Framstag    corrected forking bug
 *
 * The receive client of the sendfile package.
 * Receives, lists and deletes files from the sendfile spool.
 * 
 * Copyright  1995 Ulli Horlacher
 * This file is covered by the GNU General Public License
 */


#ifdef ULTRIX
  #define _POSIX_SOURCE
  #define S_IFMT   0170000 /* type of file */
  #define S_IFDIR  0040000 /* directory */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <time.h>
#include <utime.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "config.h"     /* various #defines */
#include "message.h"    /* information, warning and error messages */
#include "utf7.h"       /* UTF-7 coding */
#include "io.h"		/* read/write routines */
#include "string.h"     /* extended string functions */
#include "spool.h"	/* operations on files in the sendfile spool */

#ifndef AIX
  #ifndef CONVEXOS
    FILE *popen(const char *, const char *);
  #endif
  int pclose(FILE *);
  #if defined(IRIX) || defined(LINUX)
    #include <getopt.h>
  #else
    int getopt(int, char * const *, const char *);
  #endif
#endif


/* global variables */
char *lll,		/* last list line */
     *prg,		/* name of the game */
     tar_bin[MAXLEN],	/* the tar binary */
     gzip_bin[MAXLEN],	/* the gzip binary */
     recode_bin[MAXLEN],/* the recode binary */
     userspool[FLEN];	/* user spool directory */
int client=1,		/* flag to determine client or server */
    quiet=0;		/* quiet mode: no questions asked */
mode_t cmask;		/* umask value */
unsigned long blksize;	/* file system block size */


/* print short help usage text */
int usage(void);

/* list all spool files */
void list(struct senderlist *, int);

/* receive a spool file */
void receive_sf(struct filelist *, int);

/* translate NVT format file to Unix text format file */
void crlf2lf(FILE *, FILE *, const char *);

/* copy a file */
int fcopy(const char *, const char *, unsigned long, mode_t);

/* spawn a subprocess and direct output to a file */
int spawn(char **, const char *, mode_t);

/* shameless plug from GNU fileutils (getdate.c) :-) */
time_t get_date(char *, void *);  


int main(int argc, char *argv[]) 
{ extern int 
    optind;
  int 
    i,			/* simple loop count */
    number=0,		/* flag for specifying a spool file number */
    listformat=0,	/* format of listing */
    del=0,		/* flag for deleting */
    all=0,		/* flag for receiving all */
    bounce=0,		/* flag for bouncing files */
    rename=0,		/* flag for renaming files */
    id,			/* spool id */
    found=0,		/* flag for found spool id */
    opt;		/* option to test for */
  char 
    *cp,		/* a simple string pointer */
    tmp[3*MAXLEN],	/* temporary string */
    dummy[MAXLEN],	/* dummy string for utf2iso */
    fname[MAXLEN],	/* displayble file name */
    bouncelist[MAXLEN],	/* list of spool files to bounce */
    pattern[MAXLEN];	/* file name pattern to search for */
  struct stat finfo;	/* information about a file */
  struct passwd *pwe;	/* password entry */
  struct filelist *flp;	/* file list pointer */
  struct senderlist 
    *sls,		/* sender list start */
    *slp;		/* sender list pointer */
  FILE *inf;

  lll="\0";
  *bouncelist=0;
  prg=argv[0];
  cmask=umask(0);
  umask(cmask);
  
  /* no arguments? */
  if (argc==1) 
  { listformat=1;
    lll="Type \"receive -a\" to receive all files or \"receive -h\" for a short help.\n\n";
  } else 
  { 
    /* scan the command line */
    while ((opt=getopt(argc,argv,"h?lLndarbVq")) > 0) 
    { switch (opt) 
      { case ':':
        case 'h':
        case '?': exit(usage());
        case 'l': listformat=1; break;
        case 'L': listformat=2; break;
        case 'n': number=1; break;
        case 'r': rename=1; break;
        case 'b': bounce=1; break;
        case 'd': del=1; break;
        case 'a': all=1; break;
        case 'q': quiet=1; break;
        case 'V': message(prg,'I',"version "VERSION" revision "REVISION); 
	          exit(0);
      }
    }

  }

  /* too few arguments? */
  if (argc-optind==0 && (rename==1 || del==1) && all==0) exit(usage());
  if (argc-optind==0 && bounce==1 && all==1) exit(usage());
  if (argc-optind<2 && bounce==1 && all==0) exit(usage());
  
  /* get the own user name */
  if ((pwe=getpwuid(getuid())) == NULL)
    message(prg,'F',"cannot determine own user name");
  
  /* does the spool directory exist? */
  sprintf(userspool,SPOOL"/%s",pwe->pw_name);
  if (stat(userspool,&finfo)<0 || (finfo.st_mode&S_IFMT)!=S_IFDIR) 
  { sprintf(tmp,"spool directory %s does not exist",userspool);
    message(prg,'W',tmp);
    exit(1);
  }
 
  /* file system block size for copy operation */  
  blksize=finfo.st_blksize;

  /* set log file read status (st_atime) for xhoppel */
  sprintf(tmp,"%s/log",userspool);
  inf=fopen(tmp,"r");
  if (inf)
  { fgets(tmp,1,inf);
    fclose(inf);
  }
  	       
  /* are there any files to receive? */
  sls=scanspool("");
  if (sls==NULL) 
  { message(prg,'W',"no files in spool directory");
    exit(1);
  }

  /* incompatible options? */
  if (bounce)
  { if (listformat||rename||del) 
      message(prg,'W',"you cannot use any other option "
		      "when bouncing a file - ignored");
    listformat=rename=del=0;
    argc--;
  }

  /* support programs defaults */
  strcpy(tar_bin,TAR);
  strcpy(gzip_bin,GZIP);
  strcpy(recode_bin,RECODE);
  
  /* look for environment variables */
  if ((cp=getenv("SF_TAR")))	strcpy(tar_bin,cp);
  if ((cp=getenv("SF_GZIP")))	strcpy(gzip_bin,cp);
  if ((cp=getenv("SF_RECODE"))) strcpy(recode_bin,cp);

  /* do the support programs really exist? */
  if (access(tar_bin,X_OK)<0)	 strcpy(tar_bin,"tar");
  if (access(gzip_bin,X_OK)<0)	 strcpy(gzip_bin,"gzip");
  if (access(recode_bin,X_OK)<0) strcpy(recode_bin,"recode");
  
  /* list files? */
  if (listformat) 
  { list(sls,listformat);
    exit(0);    
  }

  /* number specified? */
  if (number) 
  { 
    /* loop over all args */
    for (i=optind; i<argc; i++) 
    { id=atoi(argv[i]);
      
      /* loop over sender list */
      for (slp=sls, found=0; slp!=NULL && found==0; slp=slp->next) 
      { 
	/* loop over files list */
        for (flp=slp->flist; flp!=NULL && found==0; flp=flp->next) 
	{   
          /* spool id found and spool file complete? */
	  if (flp->id==id && flp->csize==flp->tsize) 
	  {     
	    /* delete, bounce or receive spool file? */
	    if (del) 
	      delete_sf(flp,1);
	    else if (bounce)
	      sprintf(bouncelist,"%s %d",bouncelist,id);
	    else
	      receive_sf(flp,rename);
	      
	    found=1;

	  }
	}
      }
      
      /* not found? */
      if (found==0) 
      { sprintf(tmp,"spool file %d not found",id);
	message(prg,'W',tmp);
      }
      
    }

  } else { /* file name specified */

    /* loop over all args */
    for (i=optind; i<argc || all; i++) 
    { if (all)
        strcpy(pattern,"*");
      else
      	strcpy(pattern,argv[i]);
      
      /* loop over sender list */
      for (slp=sls; slp!=NULL; slp=slp->next) 
      { 
        /* loop over files list */
        for (flp=slp->flist; flp!=NULL; flp=flp->next) 
	{   
      	  /* spool file incomplete? */
          if (flp->csize!=flp->tsize) continue;
	  
	  /* translate UTF-7 name to the displayable file name */
	  utf2iso(1,dummy,fname,dummy,flp->fname);
	  
	  /* match? */
	  if (simplematch(fname,pattern,0)==1) 
	  {     
	    /* delete, bounce or receive spool file? */
            if (del) 
              delete_sf(flp,1);
	    else if (bounce)
	      sprintf(bouncelist,"%s %d",bouncelist,flp->id);
            else
	      receive_sf(flp,rename);
	    
	    found=1;
	  }
	}
      }
      
      /* not found? */
      if (found==0) 
      { sprintf(tmp,"file %s not found",pattern);
	message(prg,'W',tmp);
      }
      
      if (all) break;
    }
  }

  /* files to bounce? */
  if (bounce && *bouncelist)
  { sprintf(tmp,"sendfile -b %s %s",bouncelist,argv[argc]);
    system(tmp);
  }
  
  exit(0);    
}


/* 
 * list - list all spool files with attributes
 * 
 * INPUT:  sls  - sender list start
 */
void list(struct senderlist *sls, int format) 
{ int i;			/* simple loop counter */
  char 
    *cp1,*cp2,	    		/* simple character pointer */
    dummy[MAXLEN],		/* dummy string for utf2iso */
    line[MAXLEN],		/* line to read in */
    showtar[MAXLEN],		/* shell command to look inside tar archive */
    show_fname[MAXLEN];		/* file name to display */
  struct senderlist *slp;	/* sender list pointer */
  struct filelist *flp;		/* file list pointer */
  FILE *pipe;			/* pipe input stream */
  
  /* loop over sender list */
  for (slp=sls; slp!=NULL; slp=slp->next) 
  { 
    /* print from header */
    printf("\nFrom %s\n",slp->from);
    for (i=1; i<80 && i<=strlen(slp->from)+5; i++) printf("-");
    printf("\n");
	   
    /* loop over files list */
    for (flp=slp->flist; flp!=NULL; flp=flp->next) 
    { 
      /* not complete? */
      if (flp->csize!=flp->tsize) continue;
      
      /* print spool file informations */
      utf2iso(1,dummy,show_fname,dummy,flp->fname);
      printf("%3d) %s  %5d KB  %s",
             flp->id,flp->rdate,(flp->osize+999)/1000,show_fname);

      /* tar archive? */
      if (flp->flags&F_TAR) printf(" (archive)\n"); else printf("\n");
    
      /* comment available? */
      cp1=flp->comment;
      if (*cp1)
      { while ((cp2=strchr(cp1,'\n')))
	{ *cp2=0;
	  printf("     \"%s\"\n",cp1);
	  cp1=cp2+1;
	}
	printf("     \"%s\"\n",cp1);
      }

      /* on verbose mode look inside tar */
      if (flp->flags&F_TAR && format==2) 
      {   
	/* compressed or not compressed, this is the question! :-) */
	if (flp->flags&F_COMPRESS) 
	  sprintf(showtar,"%s -d < %s/%d.d | %s tvf -",
		  gzip_bin,userspool,flp->id,tar_bin);
	else
	  sprintf(showtar,"%s tvf %s/%d.d",tar_bin,userspool,flp->id);
	  
	/* sneak inside... */
	if ((pipe=popen(showtar,"r")) == NULL)
	  message(prg,'E',"contents of archive is not accessible");
	else
	  while (fgets(line,MAXLEN-1,pipe)) printf("     %s",line);
	pclose(pipe);

      }
    }
  }
  
  printf("\n%s",lll);
}


/* 
 * find_id - find a spool file in the list-structure
 * 
 * INPUT:  id   - spool file id number
 * 
 * RETURN: pointer to file list element, NULL if not found
 */
struct filelist *find_id(struct senderlist *sls, int id) 
{ char msg[MAXLEN];		/* information/error message */
  struct senderlist *slp;	/* sender list pointer */
  struct filelist *flp;		/* file list pointer */
  
  /* loop over sender list */
  for (slp=sls; slp!=NULL; slp=slp->next) 
  { 
    /* loop over files list */
    for (flp=slp->flist; flp!=NULL; flp=flp->next) 
    { 
      /* found file? */
      if (flp->id == id) return(flp);
    
    }
  }
  
  sprintf(msg,"cannot find spool file #%d",id);
  errno=0;
  message(prg,'E',msg);
  
  return(NULL);
}


/* 
 * receive_sf - receive a spool file
 * 
 * INPUT:  flp     - file list element
 *         rename  - flag for renaming files bevor receiving
 */
void receive_sf(struct filelist *flp, int rename) 
{ int 
    utf,			/* return code from utf2iso */
    tom=0;			/* tar overwrite message flag */
  char 
    *cp,			/* simple character pointer */
    *sad[10],			/* spawn argument descriptor */
    tmp[2*MAXLEN],		/* temporary string */
    cmd[2*MAXLEN],		/* command string for system() */
    sfile[MAXLEN],		/* spool data file */
    fname[MAXLEN],		/* file name to write */
    error[MAXLEN],		/* error log file */
    iso_name[MAXLEN],		/* file name in ISO Latin-1 */
    normal_name[MAXLEN],	/* file name in display format */
    shell_name[MAXLEN];		/* file name in shell handling format */
  static char 
    storeformat='c',		/* file name storing format */
    overwrite='n';		/* overwrite mode */
  struct stat finfo;		/* information about a file */
  struct utimbuf utb;		/* binary time */
  time_t dummy;			/* dummy arg for localtime() */
  FILE 
    *pipe,			/* pipe input stream */
    *inf,			/* spool data file to read */
    *outf;			/* file to create */
    
  /* oops? */
  if (flp==NULL) return;

  if (quiet) 
  { storeformat='C';
    overwrite='Y';
  }
  
  (void) localtime(&dummy);
  sprintf(error,"%s/.sendfile.error.%d",getenv("HOME"),(int)getpid());

  /* translate UTF-7 file name */
  utf=utf2iso(1,iso_name,normal_name,shell_name,flp->fname);

  /* tar archive? */
  if (flp->flags&F_TAR) 
  { 
    /* spool file compressed? */
    if (flp->flags&F_COMPRESS)
      sprintf(cmd,"%s -d < %s/%d.d | %s tf -",
	      gzip_bin,userspool,flp->id,tar_bin);
    else
      sprintf(cmd,"%s tf %s/%d.d",tar_bin,userspool,flp->id);
    
    /* open pipe to read tar file-info */
    if ((pipe=popen(cmd,"r")) == NULL) 
    { message(prg,'E',"cannot open spool file for reading");
      return;
    }
    
    /* loop over all files in tar archive */
    while (fgets(fname,MAXLEN-1,pipe)) 
    { 
      /* does the file already exist? */
      if ((cp=strchr(fname,'\n'))) *cp=0;
      if (overwrite!='Y' && stat(fname,&finfo)==0) 
      { if (!tom) printf("archive %s will overwrite:\n",normal_name);
        tom=1;
        printf("  %s\n",fname);
      }
    }

    pclose(pipe);
    
    /* ask user */
    if (tom && overwrite!='Y') 
    { printf("Overwrite it (yYnN)? ");
      gets(tmp);
      overwrite=tmp[0];
      if (toupper(overwrite)!='Y') return;
    }

    sprintf(tmp,"receiving from archive %s :",normal_name);
    message(prg,'I',tmp);

    /* spool file compressed? */
    if (flp->flags&F_COMPRESS)
      sprintf(cmd,"%s -d < %s/%d.d | %s xvf - 2>%s",
	      gzip_bin,userspool,flp->id,tar_bin,error);
    else
      sprintf(cmd,"%s xvf %s/%d.d 2>%s",tar_bin,userspool,flp->id,error);
    
    /* receive tar archive and check for errors */
    if (system(cmd)!=0 || (stat(error,&finfo)==0 && finfo.st_size>0)) 
    { errno=0;
      sprintf(tmp,"errors while receive %s :",normal_name);
      message(prg,'E',tmp);
      inf=fopen(error,"r");
      while (fgets(tmp,MAXLEN-1,inf)) printf("%s",tmp);
      fclose(inf);
      sprintf(tmp,"leaving %s in spool intact",normal_name);
      message(prg,'I',tmp);
      unlink(error);
      return;
    } else 
    { 
      /* delete tar spool file */
      unlink(error);
      delete_sf(flp,0);
      return;
    }
    
  }

  /* rename files before receiving? */
  if (rename)
  {  printf("Rename '%s' to? ",normal_name);
    if (gets(fname)==NULL || *fname=='\n' || *fname==0) strcpy(fname,normal_name);
    if ((cp=strrchr(fname,'\n'))) *cp=0;
  } else 
  {
    /* need user request? */
    if (!strchr("CNS",storeformat)) 
    { 
      /* found Unicode or 0x0? */
      if (utf==1) 
      { printf("The next file name contains characters which are not allowed "
	       "in Unix.\nThese characters have been substituted with '_'.\n");
      }
    
      /* found control or meta characters? */
      if (utf&2) 
        { printf("The next file name contains characters which may cause problems"
	       " with your shell or\nmay do strange things with your terminal."
	       "\nThese characters have been substituted with '_'.\n");
	if (utf&1) 
	  printf("Non-valid characters for Unix file names have been substituted"
		 ", too.\n");
      
	/* let user choose file name format */
	/*if (strcmp(normal_name,shell_name)!=0) */
	printf("(c)omplete file name with control code characters is not "
	       "displayable\n");
	printf("(n)ormal file name: >%s<\n",normal_name);
	printf("(s)hell file name:  >%s<\n",shell_name);
	do 
	{ printf("c or n may cause severe security problems! "
		 "Kids, don't try this at home!");
	  printf("Select one of c n s (or C N S for no more questions): ");
	  gets(tmp);
	  storeformat=tmp[0];
	} while (strchr("cnsCNS",storeformat)==NULL);

      }
    }

    /* set file name */
    switch (toupper(storeformat)) 
    { case 'C': strcpy(fname,iso_name); break;
      case 'N': strcpy(fname,normal_name); break;
      case 'S': strcpy(fname,shell_name); break;
    }

  }

  /* does the file already exist? */
  if (overwrite!='Y' && stat(fname,&finfo)==0) 
  { printf("%s already exists. Overwrite it (yYnN)? ",fname);
    gets(tmp);
    overwrite=tmp[0];
    if (toupper(overwrite)!='Y') return;
  }

  /* file writeable? */
  if ((outf=fopen(fname,"w")) == NULL) 
  { sprintf(tmp,"cannot open %s for writing",fname);
    message(prg,'E',tmp);
    return;
  }
  fclose(outf);
      
  /* spool file compressed? */
  if (flp->flags&F_COMPRESS) 
  { 
    /* source or text format? */
    if ((flp->flags&F_SOURCE) || (flp->flags&F_TEXT)) 
    { 
      /* open pipe to uncompress spool file */
      sprintf(cmd,"%s -d < %s/%d.d",gzip_bin,userspool,flp->id);
      if ((pipe=popen(cmd,"r")) == NULL) 
      { message(prg,'E',"cannot open spool file for reading");
        return;
     }
        
      /* open output file */
      if ((outf=fopen(fname,"w")) == NULL) 
      { sprintf(tmp,"cannot open %s for writing",fname);
        message(prg,'E',tmp);
        pclose(pipe);
        return;
      }
      
      /* translate CR LF to LF and write to output file */
      crlf2lf(pipe,outf,fname);

      /* ready */
      pclose(pipe);
      fclose(outf);

    } else {  /* binary format */
      
      /* uncompress and receive binary file */
      sprintf(sfile,"%s/%d.d",userspool,flp->id);
      sad[0]=gzip_bin;
      sad[1]="-dc";
      sad[2]=sfile;
      sad[3]=NULL;
      
      if (spawn(sad,fname,cmask)<0) 
      { errno=0;
        sprintf(tmp,"cannot receive %s",fname);
        message(prg,'E',tmp);
        return;
      }
    }
  
  } else {  /* not compressed */
    
    /* source or text format? */
    if ((flp->flags&F_SOURCE) || (flp->flags&F_TEXT)) 
    { 
      /* open input file */
      sprintf(sfile,"%s/%d.d",userspool,flp->id);
      if ((inf=fopen(sfile,"r")) == NULL) 
      { message(prg,'E',"cannot open spool file for reading");
        return;
      }
  
      /* open output file */
      if ((outf=fopen(fname,"w")) == NULL) 
      { sprintf(tmp,"cannot open %s for writing",fname);
        message(prg,'E',tmp);
        fclose(inf);
        return;
      }

      /* translate CR LF to LF and write to output file */
      crlf2lf(inf,outf,fname);

      /* ready */
      fclose(inf);
      fclose(outf);

    } else {  /* binary file */

      /* copy file */
      sprintf(tmp,"%s/%d.d",userspool,flp->id);
      if (fcopy(tmp,fname,blksize,cmask)<0) 
      { sprintf(tmp,"cannot receive %s",fname);
	errno=0;
        message(prg,'F',tmp);
      }

    }
  }

  /* delete spool file */
  delete_sf(flp,0);
  
  /* executable flag set? */
  if (flp->flags&F_EXE) chmod(fname,(S_IRWXU|S_IRWXG|S_IRWXO)&~cmask);

  /* foreign character set in text file? */
  if ((flp->flags&F_TEXT) && !streq(flp->charset,CHARSET)) 
  { 
    /* call GNU recode */
    sprintf(tmp,"%s:"CHARSET,flp->charset);
    sad[0]="recode";
    sad[1]=tmp;
    sad[2]=fname;
    sad[3]=NULL;
    if (spawn(sad,NULL,cmask)<0) 
    { sprintf(tmp,"cannot translate character set in %s",fname);
      message(prg,'E',tmp);
    }

  }

  /* set the original date */
  strcpy(tmp,flp->date);
  if (!strstr(tmp,"UTC")) strcat(tmp," UTC");
  utb.actime=utb.modtime=get_date(tmp,NULL);
  utime(fname,&utb);
  
  sprintf(tmp,"%s received",fname);
  message(prg,'I',tmp);
}

    
/* 
 * crlf2lf - translate NVT format file to Unix text format file
 * 
 * INPUT:  inf	  - file to read from
 *         outf   - file to write to
 *         fname  - file name to write
 */
void crlf2lf(FILE *inf, FILE *outf, const char *fname) 
{ int c1,c2;		/* characters to read in */
  char tmp[MAXLEN];	/* temporary string */

  /* read first char */
  c1=fgetc(inf);
  
  /* loop until EOF */
  while ((c2=fgetc(inf)) != EOF) 
  { 
    /* crlf? */
    if (c1=='\r' && c2=='\n') 
    { 
      /* write lf */
      if(fputc(c2,outf)==EOF) 
      { sprintf(tmp,"cannot write to %s",fname);
        message(prg,'E',tmp);
        return;
      }
      
      /* read next char */
      if ((c2=fgetc(inf)) == EOF) return;

    } else 
    { 
      /* write char */
      if(fputc(c1,outf)==EOF) 
      { sprintf(tmp,"cannot write to %s",fname);
        message(prg,'E',tmp);
        return;
      }
      
    }
    c1=c2;
  }
  
  /* write last char */
  if(fputc(c1,outf)==EOF) 
  { sprintf(tmp,"cannot write to %s",fname);
    message(prg,'E',tmp);
    return;
  }

}


/*
 * fcopy - copy a file
 * 
 * INPUT:  from     - source file
 *         to       - destination file
 *         blksize  - block size
 *         cmask    - umask value
 * 
 * RETURN: 0 on success, -1 on failure
 * 
 */
int fcopy(const char *from, const char *to, 
	  unsigned long blksize, mode_t cmask) 
{ int fdin, fdout;	/* file descriptor in/out */
  char tmp[MAXLEN],	/* temporary string */
       *buf;		/* copy buffer */
  int bytes;		/* read bytes */


  /* ANSI C can not dynamicly allocate with buf[blksize] */
  buf=(char *)malloc(blksize);
  if (!buf) message(prg,'F',"out of memory");
  
  /* open source file */
  fdin=open(from,O_RDONLY,0);
  if (fdin<0) 
  { sprintf(tmp,"error opening %s",from);
    message(prg,'E',tmp);
    free(buf);
    return(-1);
  }
  
  /* open destination file */
  fdout=creat(to,0666&~cmask);
  if (fdout<0) 
  { close(fdin);
    sprintf(tmp,"error creating %s",to);
    message(prg,'E',tmp);
    free(buf);
    return(-1);
  }
 
  /* read until EOF */
  while ((bytes=readn(fdin,buf,blksize)) > 0) 
  { 
    /* write to destination file */
    if (writen(fdout,buf,bytes)<0) 
    { 
      /* write error */
      close(fdin);
      close(fdout);
      sprintf(tmp,"error writing %s",to);
      message(prg,'E',tmp);
      free(buf);
      return(-1);
    }
  }
  
  free(buf);
  close(fdin);
  close(fdout);

  /* read error? */
  if (bytes<0) 
  { sprintf(tmp,"error reading %s",from);
    message(prg,'E',tmp);
    return(-1);
  }

  return(0);
}
  

/* 
 * spawn  - spawn a subprocess and direct output to a file
 * 
 * INPUT:  sad     - spawn argument descriptor
 *         output  - output file
 *         cmask   - protection mask
 * 
 * RETURN: 0 on success, -1 on failure
 * 
 */
int spawn(char **sad, const char *output, mode_t cmask) 
{ int status,		/* fork status */
      fd;		/* output file descriptor */
  char tmp[MAXLEN];	/* temporary string */
  pid_t pid;		/* process id */
  
  /* spawn subprocess */
  pid=fork();
  
  /* failed? */
  if (pid<0) 
  { message(prg,'E',"cannot fork subprocess");
    return(-1);
  }

  /* is this the subprocess? */
  if (pid==0) 
  { 
    /* redirect stdout? */
    if (output) 
    { 
      /* close stdout */
      close(1);
    
      /* open output file as stdout */
      fd=creat(output,0666&~cmask);
      if (fd!=1) 
      { message(prg,'E',"file descriptor mismatch");
	return(-1);
      }
    
    }
    
    /* execute program */
    execvp(sad[0],sad);
    
    /* oops - failed */
    exit(1);
  }
  
  /* wait for termination of subprocess */
#ifdef NEXT
  wait(&status);
#else
  waitpid(pid,&status,0);
#endif
  
  /* error in subprocess? */
  if (status) 
  { errno=0;
    sprintf(tmp,"%s failed",sad[0]);
    message(prg,'E',tmp);
    return(-1);
  }
  
  return(0);
}


/*
 * usage - print short help usage text
 */
int usage() 
{ fprintf(stderr,"usage:  %s [-dq] file-name [...]\n",prg);
  fprintf(stderr,"   or:  %s -n [-dqr] file-number [...]\n",prg);
  fprintf(stderr,"   or:  %s [-l] [-L]\n",prg);
  fprintf(stderr,"   or:  %s -a [-dqr]\n",prg);
  fprintf(stderr,"   or:  %s -b file-name [...] user[@host]\n",prg);
  fprintf(stderr,"   or:  %s -bn file-number [...] user[@host]\n",prg);
  fprintf(stderr,"   or:  %s -ba user[@host]\n",prg);
  fprintf(stderr,"options:  -l  short list of files\n");
  fprintf(stderr,"          -L  full list of files\n");
  fprintf(stderr,"          -n  specify a file number\n");
  fprintf(stderr,"          -a  specify all files\n");
  fprintf(stderr,"          -d  delete\n");
  fprintf(stderr,"          -q  quiet mode: no questions asked\n");
  fprintf(stderr,"          -r  rename file when receiving\n");
  fprintf(stderr,"          -b  bounce (forward) a file\n");
  fprintf(stderr,"examples:  receive -a         ! receive all files\n");
  fprintf(stderr,"           receive -n 1       ! receive file with number 1\n");
  fprintf(stderr,"           receive blubb      ! receive file with name blubb\n");
  return(2);
}
